from oelint_adv.cls_rule import Rule


class NoSpaceEmptyLineRule(Rule):
    def __init__(self):
        super().__init__(id="oelint.spaces.emptyline",
                         severity="warning",
                         message="Empty lines shall not contain spaces")

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if i.Raw.strip("\n") and not i.Raw.strip():
                res.append(i)
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.Raw = "\n"
            res.append(_file)
        return res



from oelint_adv.cls_rule import Rule
import re


class NoTabs(Rule):
    def __init__(self):
        super().__init__(id="oelint.tabs.notabs",
                         severity="warning",
                         message="Don't use tabs use spaces")

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if i.Raw and "\t" in i.Raw:
                res.append(i)
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.Raw = re.sub(r"\t", "    ", i.Raw)
            res.append(_file)
        return res



from oelint_adv.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskMultiAppends(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.multiappends",
                         severity="error",
                         message="Multiple appends to the same function in the same file won't work in bitbake")

    def check(self, _file, stash):
        res = []
        _stash = stash.GetItemsFor(
            filename=_file, classifier=Function.CLASSIFIER)
        for item in _stash:
            if not item.SubItem:
                continue
            if len([x for x in _stash if x.FuncName == item.FuncName and x.SubItem == item.SubItem and not x.IsFromAppend()]) > 1:
                res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_item import Function
from oelint_adv.cls_item import TaskAdd
from oelint_adv.cls_rule import Rule
from oelint_adv.const_func import FUNC_ORDER


class TaskAddNoTaskBody(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.addnotaskbody",
                         severity="warning",
                         message="The added task is not existing or has no body")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=TaskAdd.CLASSIFIER):
            if item.FuncName in FUNC_ORDER:
                # not for builtin types
                continue
            _ta = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                    attribute="FuncName", attributeValue=item.FuncName)
            if not any(_ta):
                res += self.finding(item.Origin, item.InFileLine)
            elif not any([x for x in _ta if x.FuncBodyStripped]):
                res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarMultiLineIndent(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.multilineident",
                         severity="info",
                         message="On a multiline assignment, line indent is desirable. Current {}/{}")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if not i.IsMultiLine():
                continue
            _rawclean = i.GetRawCleaned()
            _needle = i.VarValue.lstrip('"')
            if len(_needle) > 10:
                _needle = _needle[:10]
            _value = _rawclean[_rawclean.find("=") + 1:]
            _value = _value.lstrip("+ ")
            _lines = [x for x in _value.split("\\") if x]
            if any(_lines):
                _calcoffset = i.Raw.find(_lines[0])
                for _line in _lines[1:]:
                    _thisline = (len(_line) - len(_line.lstrip(" "))) - 1
                    if _thisline < _calcoffset:
                        res += self.finding(i.Origin, i.InFileLine + _lines.index(
                            _line), self.FormatMsg(_thisline, _calcoffset))
        return res



import os

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarOverride(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.override",
                         severity="error",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        __varnames = [x.VarName for x in stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)]
        for v in __varnames:
            if v == "inherit":
                # This will be done by another rule
                continue
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=v)
            items = sorted(items, key=lambda x: x.Line, reverse=False)
            if len(items) > 1:
                if all([not x.IsAppend() for x in items]):
                    self.OverrideMsg("Variable '{}' is set by {}".format(
                        v, ",".join([os.path.basename(x.Origin) for x in items])))
                    res += self.finding(items[0].Origin, items[0].InFileLine)
        return res



import textwrap


class Item():
    ATTR_LINE = "Line"
    ATTR_RAW = "Raw"
    ATTR_ORIGIN = "Origin"
    CLASSIFIER = "Item"
    ATTR_SUB = "SubItem"

    def __init__(self, origin, line, infileline, rawtext):
        self.Line = line
        self.Raw = rawtext
        self.Links = []
        self.Origin = origin
        self.InFileLine = infileline

    def extract_sub(self, name):
        chunks = name.split("_")
        if len(chunks) > 2 and chunks[-1].islower():
            return ("_".join(chunks[:-1]), chunks[-1])
        return ("_".join(chunks), "")

    def IsFromAppend(self):
        return self.Origin.endswith(".bbappend")

    def AddLink(self, _file):
        self.Links.append(_file)
        self.Links = list(set(self.Links))

    def GetAttributes(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def GetRawCleaned(self, chars="\t\n\r"):
        res = self.Raw
        for c in chars:
            res = res.replace(c, " ")
        return res

    def __repr__(self):
        return "{} -- {}\n".format(self.__class__.__name__, self.GetAttributes())


class Variable(Item):
    ATTR_VAR = "VarName"
    ATTR_VARVAL = "VarValue"
    ATTR_VARVALSTRIPPED = "VarValueStripped"
    CLASSIFIER = "Variable"
    VARIABLE_APPEND_NEEDLES = ["+="]

    def __init__(self, origin, line, infileline, rawtext, name, value):
        super().__init__(origin, line, infileline, rawtext)
        self.VarName, self.SubItem = self.extract_sub(name)
        self.VarValue = value
        self.VarValueStripped = self.VarValue.strip().lstrip('"').rstrip('"')

    def IsAppend(self):
        return any([x for x in Variable.VARIABLE_APPEND_NEEDLES if self.Raw.find(x) != -1]) or self.Raw.find("_append") != -1

    def IsMultiLine(self):
        return "\\" in self.Raw


class Comment(Item):
    CLASSIFIER = "Comment"

    def __init__(self, origin, line, infileline, rawtext):
        super().__init__(origin, line, infileline, rawtext)


class Include(Item):
    CLASSIFIER = "Include"
    ATTR_INCNAME = "IncName"

    def __init__(self, origin, line, infileline, rawtext, incname):
        super().__init__(origin, line, infileline, rawtext)
        self.IncName = incname


class Function(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_FUNCBODY = "FuncBody"
    CLASSIFIER = "Function"

    def __init__(self, origin, line, infileline, rawtext, name, body, python=False, fakeroot=False):
        super().__init__(origin, line, infileline, rawtext)
        self.IsPython = python is not None
        self.IsFakeroot = fakeroot is not None
        self.FuncName, self.SubItem = self.extract_sub(name.strip())
        if self.SubItem not in ["", "append", "remove", "class-target", "class-native"]:
            self.FuncName += "_" + self.SubItem
        self.FuncBody = body
        self.FuncBodyStripped = body.replace(
            "{", "").replace("}", "").replace("\n", "").strip()
        self.FuncBodyRaw = textwrap.dedent(rawtext[rawtext.find("{") + 1:].rstrip().rstrip("}"))


class PythonBlock(Item):
    ATTR_FUNCNAME = "FuncName"
    CLASSIFIER = "PythonBlock"

    def __init__(self, origin, line, infileline, rawtext, name):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name


class TaskAssignment(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_VAR = "VarName"
    ATTR_VARVAL = "VarValue"
    CLASSIFIER = "TaskAssignment"

    def __init__(self, origin, line, infileline, rawtext, name, ident, value):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name
        self.VarName = ident
        self.VarValue = value


class TaskAdd(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_BEFORE = "Before"
    ATTR_AFTER = "After"
    CLASSIFIER = "TaskAdd"

    def __init__(self, origin, line, infileline, rawtext, name, before="", after=""):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name
        self.Before = [x for x in (before or "").split(" ") if x]
        self.After = [x for x in (after or "").split(" ") if x]



import importlib
import inspect
import os
import pkgutil


class Rule():
    def __init__(self, id="", severity="", message=""):
        self.ID = id
        self.Severity = severity
        self.Msg = message

    def check(self, _file, stash):
        return []

    def fix(self, _file, stash):
        return []

    def finding(self, _file, _line, override_msg=None):
        return ["{}:{}:{}:{}:{}".format(os.path.abspath(_file), _line, self.Severity, self.ID, override_msg or self.Msg)]

    def __repr__(self):
        return "{}".format(self.ID)

    def FormatMsg(self, *args):
        return self.Msg.format(*args)

    def OverrideMsg(self, newmsg):
        self.Msg = newmsg


def load_rules(add_rules=[]):
    res = []
    _path_list = {
        "base": os.path.dirname(os.path.abspath(__file__))
    }
    for ar in add_rules:
        _path_list[ar] = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "rule_{}".format(ar))
    for k, v in _path_list.items():
        packages = pkgutil.walk_packages(path=[v])
        for importer, name, is_package in packages:
            if k != "base":
                name = os.path.basename(v) + "." + name
            mod = importlib.import_module(name)
            for m in inspect.getmembers(mod, inspect.isclass):
                try:
                    if issubclass(m[1], Rule):
                        inst = m[1]()
                        if inst.ID:
                            res.append(inst)
                except Exception as ex:
                    print("{} -> {}".format(name, ex))
    return res



import glob
import os
from urllib.parse import urlparse

from oelint_adv.cls_item import Variable


def get_files(stash, _file, pattern):
    res = []
    src_uris = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                 attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
    files_paths = list(
        set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern) for x in src_uris]))
    for item in src_uris:
        files_paths += list(set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern)
                                 for x in stash.GetItemsFor(filename=item.Origin)]))
    for item in files_paths:
        res += glob.glob(item)
    return list(set(res))


def get_scr_components(string):
    """
        Parses an URL string
        returns a dict with
            scheme = protcol used
            src = path to call
            options = dict with options added to URL
    """
    _url = urlparse(string)
    _scheme = _url.scheme
    _options = _url.netloc.split(";")[1:]
    _path = _url.netloc.split(";")[0]
    if _url.path:
        if not _path.endswith("/") and not _url.path.startswith("/"):
            _path += "/"
        _path += _url.path
    _parsed_opt = {x.split("=")[0]: x.split("=")[1] for x in _options}
    return {"scheme": _scheme, "src": _path, "options": _parsed_opt}



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSpacesOnAssignment(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.spacesassignment",
                         severity="warning",
                         message="Suggest spaces around variable assignment. E.g. 'FOO = \"BAR\"'")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == "inherit":
                continue
            needles = [" = ", " ?= ", " ??= ", " := ", " += "]
            if not any([x for x in needles if i.Raw.find(x) != -1]):
                res += self.finding(i.Origin, i.InFileLine)
        return res




MANDATORY_VARS = [
    "SUMMARY",
    "DESCRIPTION",
    "AUTHOR",
    "HOMEPAGE",
    "SECTION",
    "LICENSE",
    "SRC_URI"
]

SUGGESTED_VARS = [
    "BUGTRACKER",
    "BBCLASSEXTEND",
    "CVE_PRODUCT"
]

VAR_ORDER = [
    "SUMMARY",
    "DESCRIPTION",
    "AUTHOR",
    "HOMEPAGE",
    "BUGTRACKER",
    "SECTION",
    "LICENSE",
    "LIC_FILES_CHKSUM",
    "DEPENDS",
    "PROVIDES",
    "PV",
    "SRC_URI",
    "SRCREV",
    "S",
    "inherit",
    "PACKAGECONFIG",
    "EXTRA_QMAKEVARS_POST",
    "EXTRA_OECONF",
    "PACKAGE_ARCH",
    "PACKAGES",
    "FILES_${PN}",
    "RDEPENDS_${PN}",
    "RRECOMMENDS_${PN}",
    "RSUGGESTS_${PN}",
    "RPROVIDES_${PN}",
    "RCONFLICTS_${PN}",
    "BBCLASSEXTEND"
]



import re

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarMultiInherit(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.multiinherit",
                         severity="warning",
                         message="'{INH}' is included multiple times")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="inherit")
        keys = []
        for i in items:
            for y in [x.strip() for x in re.split(r"\s|,", i.VarValue) if x]:
                if y not in keys:
                    keys.append(y)
                else:
                    res += self.finding(i.Origin, i.InFileLine,
                                        self.Msg.replace("{INH}", y))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import get_scr_components


class VarSRCURIWildcard(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.srcuriwildcard",
                         severity="error",
                         message="'SRC_URI' should not contain any wildcards")

    def check(self, _file, stash):
        res = []
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
        for i in _items:
            for f in [x for x in i.VarValueStripped.split(" ") if x]:
                components = get_scr_components(f)
                if components["scheme"] == "file":
                    if any([x for x in ["*"] if x in components["src"]]):
                        res += self.finding(i.Origin, i.InFileLine)
        return res



import argparse
import os
import sys

from oelint_adv.cls_rule import load_rules
from oelint_adv.cls_stash import Stash

sys.path.append(os.path.abspath(os.path.join(__file__, "..")))


def create_argparser():
    parser = argparse.ArgumentParser(
        description='Advanced OELint - Check bitbake recipes against OECore styleguide')
    parser.add_argument("--suppress", default=[],
                        action="append", help="Rules to suppress")
    parser.add_argument("--output", default=sys.stderr,
                        help="Where to flush the findings (default: stderr)")
    parser.add_argument("--fix", action="store_true", default=False,
                        help="Automatically try to fix the issues")
    parser.add_argument("--nobackup", action="store_true", default=False,
                        help="Don't create backup file when auto fixing")
    parser.add_argument("--addrules", nargs="+", default=[],
                        help="Additional non-default rulessets to add")
    parser.add_argument("files", nargs='+', help="File to parse")
    return parser


if __name__ == '__main__':
    args = create_argparser().parse_args()
    rules = [x for x in load_rules(
        add_rules=args.addrules) if str(x) not in args.suppress]
    print("Loaded rules: {}".format(",".join(sorted([str(x) for x in rules]))))
    stash = Stash()
    issues = []
    fixedfiles = []
    for f in args.files:
        try:
            stash.AddFile(f)
        except FileNotFoundError:
            print("Can't open/read {}".format(f))

    for f in stash.GetRecipes():
        for r in rules:
            if args.fix:
                fixedfiles += r.fix(f, stash)
            issues += r.check(f, stash)
    fixedfiles = list(set(fixedfiles))
    for f in fixedfiles:
        items = sorted(stash.GetItemsFor(filename=f), key=lambda x: x.Line)
        if not args.nobackup:
            os.rename(f, f + ".bak")
        with open(f, "w") as o:
            o.write("".join([x.Raw for x in items]))
            print("{}:{}:{}".format(os.path.abspath(f),
                                    "debug", "Applied automatic fixes"))

    issues = list(set(issues))
    if args.output != sys.stderr:
        args.output = open(args.output, "w")
    args.output.write("\n".join(sorted(issues)) + "\n")
    sys.exit(len(issues))



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import get_scr_components


class VarBugtrackerIsUrl(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.bugtrackerisurl",
                         severity="warning",
                         message="'BUGTRACKER' should be an URL")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="BUGTRACKER")
        for i in items:
            val = i.VarValueStripped
            try:
                result = get_scr_components(val)
                if not result["scheme"] or not result["src"]:
                    raise Exception()
            except Exception:
                res += self.finding(i.Origin, i.InFileLine)
        return res



import ast

from oelint_adv.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskPythonPrefix(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.pythonprefix",
                         severity="warning",
                         message="Tasks containing python code, should be prefixed with python in function header")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            try:
                print(item.FuncBodyRaw)
                ast.parse(item.FuncBodyRaw, "tempfile")
                if not item.IsPython:
                    res += self.finding(item.Origin, item.InFileLine)
            except Exception:
                pass
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.const_vars import SUGGESTED_VARS


class VarSuggestedExists(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.suggestedvar",
                         severity="info",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        for var in SUGGESTED_VARS:
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                self.OverrideMsg("Variable '{}' should be set".format(var))
                res += self.finding(_file, 0)
        return res



import collections
import os
import re

from oelint_adv.cls_item import Comment
from oelint_adv.cls_item import Function
from oelint_adv.cls_item import Include
from oelint_adv.cls_item import Item
from oelint_adv.cls_item import PythonBlock
from oelint_adv.cls_item import TaskAdd
from oelint_adv.cls_item import TaskAssignment
from oelint_adv.cls_item import Variable


def prepare_lines(_file, lineOffset=0):
    __func_start_regexp__ = r".*(((?P<py>python)|(?P<fr>fakeroot))\s*)*(?P<func>[\w\.\-\+\{\}\$]+)?\s*\(\s*\)\s*\{"
    with open(_file) as i:
        prep_lines = []
        _iter = enumerate(i.readlines())
        for num, line in _iter:
            raw_line = line
            if raw_line.find("\\\n") != -1:
                _, line = _iter.__next__()
                while line.find("\\\n") != -1:
                    raw_line += line
                    _, line = _iter.__next__()
                raw_line += line
            elif re.match(__func_start_regexp__, raw_line):
                _, line = _iter.__next__()
                while line.strip() != "}":
                    raw_line += line
                    _, line = _iter.__next__()
                if line.strip() == "}":
                    raw_line += line
            elif raw_line.strip().startswith("def "):
                _, line = _iter.__next__()
                while (line.startswith(" ") or line.startswith("\t")):
                    raw_line += line
                    _, line = _iter.__next__()
                    if not line.strip():
                        break
            prep_lines.append({"line": num + 1 + lineOffset, "raw": raw_line,
                               "cnt": raw_line.replace("\n", "").replace("\\", "")})
        # print(prep_lines)
        return prep_lines
    return []


def get_items(stash, _file, lineOffset=0):
    res = []
    __regex_var = r"^.*?(?P<varname>([A-Z0-9a-z_-]|\$|\{|\})+)(\s*|\t*)(\+|\?)*=(\s*|\t*)(?P<varval>.*)"
    __regex_func = r"^((?P<py>python)\s+|(?P<fr>fakeroot\s+))*(?P<func>[\w\.\-\+\{\}\$]+)?\s*\(\s*\)\s*\{(?P<funcbody>.*)\s*\}"
    __regex_inherit = r"^.*?inherit(\s+|\t+)(?P<inhname>.+)"
    __regex_comments = r"^.*?#+(?P<body>.*)"
    __regex_python = r"^(\s*|\t*)def(\s+|\t+)(?P<funcname>[a-z0-9_]+)(\s*|\t*)\:.+"
    __regex_include = r"^(\s*|\t*)(include|require)(\s+|\t+)(?P<incname>[A-za-z0-9\-\.]+)"
    __regex_addtask = r"^(\s*|\t*)addtask\s+(?P<func>\w+)\s*((before\s*(?P<before>((.*(?=after))|(.*))))|(after\s*(?P<after>((.*(?=before))|(.*)))))*"
    __regex_taskass = r"^(\s*|\t*)(?P<func>\w+)\[(?P<ident>\w+)\](\s+|\t+)=(\s+|\t+)(?P<varval>.*)"

    _order = collections.OrderedDict([
        ("comment", __regex_comments),
        ("func", __regex_func),
        ("inherit", __regex_inherit),
        ("python", __regex_python),
        ("include", __regex_include),
        ("addtask", __regex_addtask),
        ("taskassign", __regex_taskass),
        ("vars", __regex_var)
    ])

    for line in prepare_lines(_file, lineOffset):
        good = False
        for k, v in _order.items():
            m = re.match(v, line["cnt"], re.MULTILINE)
            if m:
                if k == "python":
                    if any(res) and isinstance(res[-1], PythonBlock):
                        res[-1].Raw += line["raw"]
                    else:
                        res.append(PythonBlock(
                            _file, line["line"], line["line"] - lineOffset, line["raw"], m.group("funcname")))
                elif k == "vars":
                    res.append(Variable(
                        _file, line["line"], line["line"] - lineOffset, line["raw"], m.group("varname"), m.group("varval")))
                elif k == "func":
                    res.append(Function(
                        _file, line["line"], line["line"] - lineOffset, line["raw"],
                        m.group("func"), m.group("funcbody"),
                        m.group("py"), m.group("fr")))
                elif k == "comment":
                    res.append(
                        Comment(_file, line["line"], line["line"] - lineOffset, line["raw"]))
                elif k == "inherit":
                    res.append(Variable(
                        _file, line["line"], line["line"] - lineOffset, line["raw"], "inherit", m.group("inhname")))
                elif k == "taskassign":
                    res.append(TaskAssignment(_file, line["line"], line["line"] - lineOffset, line["raw"], m.group(
                        "func"), m.group("ident"), m.group("varval")))
                elif k == "addtask":
                    _g = m.groupdict()
                    if "before" in _g.keys():
                        _b = _g["before"]
                    else:
                        _b = ""
                    if "after" in _g.keys():
                        _a = _g["after"]
                    else:
                        _a = ""
                    res.append(TaskAdd(
                        _file, line["line"], line["line"] - lineOffset, line["raw"], m.group("func"), _b, _a))
                elif k == "include":
                    tmp = stash.AddFile(os.path.abspath(os.path.join(os.path.dirname(
                        _file), m.group("incname"))), lineOffset=line["line"], forcedLink=_file)
                    res.append(Include(
                        _file, line["line"], line["line"] - lineOffset, line["raw"], m.group("incname")))
                    if any(tmp):
                        lineOffset += max([x.InFileLine for x in tmp])
                good = True
            if good:
                break
        if not good:
            res.append(
                Item(_file, line["line"], line["line"] - lineOffset, line["raw"]))
    return res
